@@ -1,38 +1,12 @@
-/****************************************************************************
-**
-** Copyright (C) 2008-2012 NVIDIA Corporation.
-** Copyright (C) 2019 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of Qt Quick 3D.
-**
-** $QT_BEGIN_LICENSE:GPL$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 or (at your option) any later version
-** approved by the KDE Free Qt Foundation. The licenses are as published by
-** the Free Software Foundation and appearing in the file LICENSE.GPL3
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2008-2012 NVIDIA Corporation.
+// Copyright (C) 2019 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
 #include "qssgvertexpipelineimpl_p.h"
 
 #include <QtQuick3DRuntimeRender/private/qssgrenderer_p.h>
 #include <QtQuick3DRuntimeRender/private/qssgrenderlight_p.h>
-#include <QtQuick3DRuntimeRender/private/qssgrendercontextcore_p.h>
+#include "../qssgrendercontextcore.h"
 #include <QtQuick3DRuntimeRender/private/qssgrendershadercache_p.h>
 #include <QtQuick3DRuntimeRender/private/qssgrendershaderlibrarymanager_p.h>
 #include <QtQuick3DRuntimeRender/private/qssgrendershadercodegenerator_p.h>
@@ -41,61 +15,69 @@
 
 QT_BEGIN_NAMESPACE
 
-QSSGMaterialVertexPipeline::QSSGMaterialVertexPipeline(const QSSGRef<QSSGProgramGenerator> &programGen,
+QSSGMaterialVertexPipeline::QSSGMaterialVertexPipeline(QSSGProgramGenerator &programGen,
                                                        const QSSGShaderDefaultMaterialKeyProperties &materialProperties,
-                                                       QSSGShaderMaterialAdapter *materialAdapter,
-                                                       QSSGDataView<QMatrix4x4> boneGlobals,
-                                                       QSSGDataView<QMatrix3x3> boneNormals,
-                                                       QSSGDataView<float> morphWeights)
-    : m_programGenerator(programGen)
+                                                       QSSGShaderMaterialAdapter *materialAdapter)
+    : m_programGenerator(&programGen)
     , defaultMaterialShaderKeyProperties(materialProperties)
     , materialAdapter(materialAdapter)
-    , boneGlobals(boneGlobals)
-    , boneNormals(boneNormals)
-    , morphWeights(morphWeights)
     , hasCustomShadedMain(false)
     , skipCustomFragmentSnippet(false)
 {
-    m_hasSkinning = boneGlobals.size() > 0;
-    m_hasMorphing = morphWeights.size() > 0;
 }
 
-static inline void insertProcessorArgs(QByteArray &snippet, const char *argKey, const char* (*argListFunc)())
+static inline void insertProcessorArgs(QByteArray &snippet, const char *argKey, const char* (*argListFunc)(), bool usesShared = false, bool isSharedInout = false)
 {
     const int argKeyLen = int(strlen(argKey));
     const int argKeyPos = snippet.indexOf(argKey);
-    if (argKeyPos >= 0)
-        snippet = snippet.left(argKeyPos) + argListFunc() + snippet.mid(argKeyPos + argKeyLen);
+    if (argKeyPos >= 0) {
+        if (!usesShared) {
+            snippet = snippet.left(argKeyPos) + argListFunc() + snippet.mid(argKeyPos + argKeyLen);
+        } else {
+            const char *inoutString = isSharedInout ? ", inout " : ", in ";
+            snippet = snippet.left(argKeyPos) + argListFunc() + inoutString + QByteArrayLiteral("QT_SHARED_VARS SHARED") + snippet.mid(argKeyPos + argKeyLen);
+        }
+    }
 }
 
-static inline void insertDirectionalLightProcessorArgs(QByteArray &snippet)
+static inline void insertDirectionalLightProcessorArgs(QByteArray &snippet, bool usesShared)
 {
-    insertProcessorArgs(snippet, "/*%QT_ARGS_DIRECTIONAL_LIGHT%*/", QSSGMaterialShaderGenerator::directionalLightProcessorArgumentList);
+    insertProcessorArgs(snippet, "/*%QT_ARGS_DIRECTIONAL_LIGHT%*/", QSSGMaterialShaderGenerator::directionalLightProcessorArgumentList, usesShared, true);
 }
 
-static inline void insertPointLightProcessorArgs(QByteArray &snippet)
+static inline void insertPointLightProcessorArgs(QByteArray &snippet, bool usesShared)
 {
-    insertProcessorArgs(snippet, "/*%QT_ARGS_POINT_LIGHT%*/", QSSGMaterialShaderGenerator::pointLightProcessorArgumentList);
+    insertProcessorArgs(snippet, "/*%QT_ARGS_POINT_LIGHT%*/", QSSGMaterialShaderGenerator::pointLightProcessorArgumentList, usesShared, true);
 }
 
-static inline void insertSpotLightProcessorArgs(QByteArray &snippet)
+static inline void insertSpotLightProcessorArgs(QByteArray &snippet, bool usesShared)
 {
-    insertProcessorArgs(snippet, "/*%QT_ARGS_SPOT_LIGHT%*/", QSSGMaterialShaderGenerator::spotLightProcessorArgumentList);
+    insertProcessorArgs(snippet, "/*%QT_ARGS_SPOT_LIGHT%*/", QSSGMaterialShaderGenerator::spotLightProcessorArgumentList, usesShared, true);
 }
 
-static inline void insertAmbientLightProcessorArgs(QByteArray &snippet)
+static inline void insertAmbientLightProcessorArgs(QByteArray &snippet, bool usesShared)
 {
-    insertProcessorArgs(snippet, "/*%QT_ARGS_AMBIENT_LIGHT%*/", QSSGMaterialShaderGenerator::ambientLightProcessorArgumentList);
+    insertProcessorArgs(snippet, "/*%QT_ARGS_AMBIENT_LIGHT%*/", QSSGMaterialShaderGenerator::ambientLightProcessorArgumentList, usesShared, true);
 }
 
-static inline void insertSpecularLightProcessorArgs(QByteArray &snippet)
+static inline void insertIblProbeProcessorArgs(QByteArray &snippet, bool usesShared)
 {
-    insertProcessorArgs(snippet, "/*%QT_ARGS_SPECULAR_LIGHT%*/", QSSGMaterialShaderGenerator::specularLightProcessorArgumentList);
+    insertProcessorArgs(snippet, "/*%QT_ARGS_IBL_PROBE%*/", QSSGMaterialShaderGenerator::iblProbeProcessorArgumentList, usesShared, true);
 }
 
-static inline void insertFragmentMainArgs(QByteArray &snippet)
+static inline void insertSpecularLightProcessorArgs(QByteArray &snippet, bool usesShared)
 {
-    insertProcessorArgs(snippet, "/*%QT_ARGS_MAIN%*/", QSSGMaterialShaderGenerator::shadedFragmentMainArgumentList);
+    insertProcessorArgs(snippet, "/*%QT_ARGS_SPECULAR_LIGHT%*/", QSSGMaterialShaderGenerator::specularLightProcessorArgumentList, usesShared, true);
+}
+
+static inline void insertFragmentMainArgs(QByteArray &snippet, bool usesShared = false)
+{
+    insertProcessorArgs(snippet, "/*%QT_ARGS_MAIN%*/", QSSGMaterialShaderGenerator::shadedFragmentMainArgumentList, usesShared, true);
+}
+
+static inline void insertPostProcessorArgs(QByteArray &snippet, bool usesShared)
+{
+    insertProcessorArgs(snippet, "/*%QT_ARGS_POST_PROCESS%*/", QSSGMaterialShaderGenerator::postProcessorArgumentList, usesShared, false);
 }
 
 static inline void insertVertexMainArgs(QByteArray &snippet)
@@ -117,8 +99,8 @@ static inline const char *customMainCallWithArguments(bool usesInstancing)
 }
 
 void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMaterialKey &inKey,
-                                                       const ShaderFeatureSetList &inFeatureSet,
-                                                       const QSSGRef<QSSGShaderLibraryManager> &shaderLibraryManager)
+                                                       const QSSGShaderFeatures &inFeatureSet,
+                                                       QSSGShaderLibraryManager &shaderLibraryManager)
 {
     QSSGShaderGeneratorStageFlags theStages(QSSGProgramGenerator::defaultFlags());
     programGenerator()->beginProgram(theStages);
@@ -131,6 +113,8 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
                 QSSGShaderKeyVertexAttribute::TexCoord0, inKey);
     const bool meshHasTexCoord1 = defaultMaterialShaderKeyProperties.m_vertexAttributes.getBitValue(
                 QSSGShaderKeyVertexAttribute::TexCoord1, inKey);
+    const bool meshHasTexCoordLightmap = defaultMaterialShaderKeyProperties.m_vertexAttributes.getBitValue(
+                QSSGShaderKeyVertexAttribute::TexCoordLightmap, inKey);
     const bool meshHasTangents = defaultMaterialShaderKeyProperties.m_vertexAttributes.getBitValue(
                 QSSGShaderKeyVertexAttribute::Tangent, inKey);
     const bool meshHasBinormals = defaultMaterialShaderKeyProperties.m_vertexAttributes.getBitValue(
@@ -144,7 +128,11 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
     const bool usesInvProjectionMatrix = defaultMaterialShaderKeyProperties.m_usesInverseProjectionMatrix.getValue(inKey);
     const bool usesPointsTopology = defaultMaterialShaderKeyProperties.m_usesPointsTopology.getValue(inKey);
     const bool usesFloatJointIndices = defaultMaterialShaderKeyProperties.m_usesFloatJointIndices.getValue(inKey);
+    const bool blendParticles = defaultMaterialShaderKeyProperties.m_blendParticles.getValue(inKey);
     usesInstancing = defaultMaterialShaderKeyProperties.m_usesInstancing.getValue(inKey);
+    m_hasSkinning = defaultMaterialShaderKeyProperties.m_boneCount.getValue(inKey) > 0;
+    const auto morphSize = defaultMaterialShaderKeyProperties.m_targetCount.getValue(inKey);
+    m_hasMorphing = morphSize > 0;
 
     vertexShader.addIncoming("attr_pos", "vec3");
     if (usesInstancing) {
@@ -154,6 +142,14 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
         vertexShader.addIncoming("qt_instanceColor", "vec4");
         vertexShader.addIncoming("qt_instanceData", "vec4");
     }
+    if (blendParticles) {
+        vertexShader.addInclude("particles.glsllib");
+        vertexShader.addUniform("qt_particleTexture", "sampler2D");
+        vertexShader.addUniform("qt_countPerSlice", "uint");
+        vertexShader.addUniform("qt_oneOverParticleImageSize", "vec2");
+        vertexShader.addUniform("qt_particleMatrix", "mat4");
+        vertexShader.addUniform("qt_particleIndexOffset", "uint");
+    }
 
     if (m_hasSkinning && meshHasJointsAndWeights) {
         vertexShader.addInclude("skinanim.glsllib");
@@ -163,13 +159,16 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
             vertexShader.addIncoming("attr_joints", "ivec4");
         vertexShader.addIncoming("attr_weights", "vec4");
 
-        vertexShader.addUniformArray("qt_boneTransforms", "mat4", boneGlobals.mSize);
-        vertexShader.addUniformArray("qt_boneNormalTransforms", "mat3", boneNormals.mSize);
+        vertexShader.addUniform("qt_boneTexture", "sampler2D");
+    }
+    if (m_hasMorphing) {
+        vertexShader.addInclude("morphanim.glsllib");
+        vertexShader.addUniformArray("qt_morphWeights", "float", morphSize);
+        vertexShader.addUniform("qt_morphTargetTexture", "sampler2DArray");
     }
-    if (m_hasMorphing)
-        vertexShader.addUniformArray("qt_morphWeights", "float", morphWeights.mSize);
 
     const bool hasCustomVertexShader = materialAdapter->hasCustomShaderSnippet(QSSGShaderCache::ShaderType::Vertex);
+    const bool hasCustomFragmentShader = materialAdapter->hasCustomShaderSnippet(QSSGShaderCache::ShaderType::Fragment);
     if (hasCustomVertexShader) {
         QByteArray snippet = materialAdapter->customShaderSnippet(QSSGShaderCache::ShaderType::Vertex,
                                                                   shaderLibraryManager);
@@ -182,10 +181,18 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
             else
                 insertVertexMainArgs(snippet);
 
-            if (m_hasSkinning) {
+            if (materialAdapter->usesCustomSkinning()) {
                 vertexShader.addInclude("skinanim.glsllib");
-                vertexShader.addUniformArray("qt_boneTransforms", "mat4", boneGlobals.mSize);
-                vertexShader.addUniformArray("qt_boneNormalTransforms", "mat3", boneNormals.mSize);
+                vertexShader.addUniform("qt_boneTexture", "sampler2D");
+                m_hasSkinning = false;
+            }
+
+            if (materialAdapter->usesCustomMorphing()) {
+                vertexShader.addInclude("morphanim_custom.glsllib");
+                if (morphSize > 0)
+                    vertexShader.addUniformArray("qt_morphWeights", "float", morphSize);
+                vertexShader.addUniform("qt_morphTargetTexture", "sampler2DArray");
+                m_hasMorphing = false;
             }
 
             if (!materialAdapter->isUnshaded()) {
@@ -206,11 +213,16 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
     vertexShader.append("    vec3 qt_vertNormal = vec3(0.0);");
     vertexShader.append("    vec3 qt_vertTangent = vec3(0.0);");
     vertexShader.append("    vec3 qt_vertBinormal = vec3(0.0);");
-    vertexShader.append("    vec2 qt_vertUV0 = vec2(0.0);");
-    vertexShader.append("    vec2 qt_vertUV1 = vec2(0.0);");
-    vertexShader.append("    ivec4 qt_vertJoints = ivec4(0);");
-    vertexShader.append("    vec4 qt_vertWeights = vec4(0.0);");
-    vertexShader.append("    vec4 qt_vertColor = vec4(1.0);"); // must be 1,1,1,1 to not alter when multiplying with it
+    if (meshHasTexCoord0 || hasCustomVertexShader)
+        vertexShader.append("    vec2 qt_vertUV0 = vec2(0.0);");
+    if (meshHasTexCoord1 || hasCustomVertexShader)
+        vertexShader.append("    vec2 qt_vertUV1 = vec2(0.0);");
+    if (m_hasSkinning || hasCustomVertexShader)
+        vertexShader.append("    ivec4 qt_vertJoints = ivec4(0);");
+    if (meshHasJointsAndWeights || m_hasSkinning || hasCustomVertexShader)
+        vertexShader.append("    vec4 qt_vertWeights = vec4(0.0);");
+    if (meshHasColors || usesInstancing || blendParticles || hasCustomVertexShader || hasCustomFragmentShader)
+        vertexShader.append("    vec4 qt_vertColor = vec4(1.0);"); // must be 1,1,1,1 to not alter when multiplying with it
 
     if (!usesInstancing) {
         vertexShader.addUniform("qt_modelViewProjection", "mat4");
@@ -221,13 +233,19 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
         vertexShader.addUniform("qt_viewProjectionMatrix", "mat4");
     }
 
+    // The custom fragment main should be skipped if this is a
+    // depth pass, but not if it is also a OpaqueDepthPrePass
+    // because then we need to know the real alpha values
     skipCustomFragmentSnippet = false;
-    for (const auto &feature : inFeatureSet) {
-        if (feature.name == QSSGShaderDefines::asString(QSSGShaderDefines::DepthPass))
-            skipCustomFragmentSnippet = feature.enabled;
-    }
+    const bool isDepthPass = inFeatureSet.isSet(QSSGShaderFeatures::Feature::DepthPass);
+    const bool isOpaqueDepthPrePass = inFeatureSet.isSet(QSSGShaderFeatures::Feature::OpaqueDepthPrePass);
+    skipCustomFragmentSnippet = (isDepthPass && !isOpaqueDepthPrePass);
 
-    if (hasCustomVertexShader) { // this is both for unshaded and shaded
+    if (hasCustomVertexShader || hasCustomFragmentShader) {
+        // This is both for unshaded and shaded. Regardless of any other
+        // condition we have to ensure the keywords (VIEW_MATRIX etc.) promised
+        // by the documentation are available in *both* the custom vertex and
+        // fragment shader snippets, even if only one of them is present.
         vertexShader.addUniform("qt_viewProjectionMatrix", "mat4");
         vertexShader.addUniform("qt_modelMatrix", "mat4");
         vertexShader.addUniform("qt_viewMatrix", "mat4");
@@ -253,6 +271,10 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
         vertexShader.append("    qt_vertUV1 = attr_uv1;");
         vertexShader.addIncoming("attr_uv1", "vec2");
     }
+    if (meshHasTexCoordLightmap) {
+        vertexShader.append("    vec2 qt_vertLightmapUV = attr_lightmapuv;");
+        vertexShader.addIncoming("attr_lightmapuv", "vec2");
+    }
     if (meshHasTangents) {
         vertexShader.append("    qt_vertTangent = attr_textan;");
         vertexShader.addIncoming("attr_textan", "vec3");
@@ -266,7 +288,7 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
         vertexShader.addIncoming("attr_color", "vec4");
     }
 
-    if (meshHasJointsAndWeights) {
+    if (meshHasJointsAndWeights && (m_hasSkinning || hasCustomVertexShader)) {
         if (usesFloatJointIndices) {
             vertexShader.addIncoming("attr_joints", "vec4");
             vertexShader.append("    qt_vertJoints = ivec4(attr_joints);");
@@ -281,7 +303,10 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
     if (usesInstancing) {
         vertexShader.append("    qt_vertColor *= qt_instanceColor;");
         vertexShader.append("    mat4 qt_instanceMatrix = mat4(qt_instanceTransform0, qt_instanceTransform1, qt_instanceTransform2, vec4(0.0, 0.0, 0.0, 1.0));");
-        vertexShader.append("    mat4 qt_instancedModelMatrix =  qt_parentMatrix * transpose(qt_instanceMatrix) * qt_modelMatrix;");
+        if (m_hasSkinning)
+            vertexShader.append("    mat4 qt_instancedModelMatrix =  qt_parentMatrix * transpose(qt_instanceMatrix);");
+        else
+            vertexShader.append("    mat4 qt_instancedModelMatrix =  qt_parentMatrix * transpose(qt_instanceMatrix) * qt_modelMatrix;");
         vertexShader.append("    mat3 qt_instancedNormalMatrix = mat3(transpose(inverse(qt_instancedModelMatrix)));");
         vertexShader.append("    mat4 qt_instancedMVPMatrix = qt_viewProjectionMatrix * qt_instancedModelMatrix;");
     }
@@ -294,11 +319,17 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
             vertexShader.append(customMainCallWithArguments(usesInstancing));
 
         if (m_hasMorphing && !hasCustomVertexShader)
-            vertexShader.append("    qt_vertPosition.xyz = qt_getMorphPosition(qt_vertPosition.xyz);");
+            vertexShader.append("    qt_vertPosition.xyz = qt_getTargetPosition(qt_vertPosition.xyz);");
 
         if (m_hasSkinning) {
-            vertexShader.append("    if (qt_vertWeights != vec4(0.0))");
-            vertexShader.append("        qt_vertPosition = qt_getSkinMatrix(qt_vertJoints, qt_vertWeights) * qt_vertPosition;");
+            vertexShader.append("    mat4 skinMat = mat4(1);");
+            vertexShader.append("    if (qt_vertWeights != vec4(0.0)) {");
+            vertexShader.append("        skinMat = qt_getSkinMatrix(qt_vertJoints, qt_vertWeights);");
+            vertexShader.append("        qt_vertPosition = skinMat * qt_vertPosition;");
+            vertexShader.append("    }");
+        }
+        if (blendParticles) {
+            vertexShader.append("    qt_vertPosition.xyz = qt_applyParticle(qt_vertPosition.xyz, qt_vertNormal, qt_vertColor, qt_vertNormal, qt_vertColor, qt_particleMatrix);");
         }
 
         if (!hasCustomShadedMain || !overridesPosition) {
@@ -315,7 +346,7 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
     } // with a custom vertex shader it is up to it to set gl_PointSize (aka POINT_SIZE)
 }
 
-void QSSGMaterialVertexPipeline::beginFragmentGeneration(const QSSGRef<QSSGShaderLibraryManager> &shaderLibraryManager)
+void QSSGMaterialVertexPipeline::beginFragmentGeneration(QSSGShaderLibraryManager &shaderLibraryManager)
 {
     fragment().addUniform("qt_material_properties", "vec4");
     fragment().addUniform("qt_rhi_properties", "vec4");
@@ -324,12 +355,15 @@ void QSSGMaterialVertexPipeline::beginFragmentGeneration(const QSSGRef<QSSGShade
         QByteArray snippet = materialAdapter->customShaderSnippet(QSSGShaderCache::ShaderType::Fragment,
                                                                   shaderLibraryManager);
         if (!materialAdapter->isUnshaded()) {
-            insertAmbientLightProcessorArgs(snippet);
-            insertSpecularLightProcessorArgs(snippet);
-            insertSpotLightProcessorArgs(snippet);
-            insertPointLightProcessorArgs(snippet);
-            insertDirectionalLightProcessorArgs(snippet);
-            insertFragmentMainArgs(snippet);
+            const bool usesShared = materialAdapter->usesSharedVariables();
+            insertAmbientLightProcessorArgs(snippet, usesShared);
+            insertIblProbeProcessorArgs(snippet, usesShared);
+            insertSpecularLightProcessorArgs(snippet, usesShared);
+            insertSpotLightProcessorArgs(snippet, usesShared);
+            insertPointLightProcessorArgs(snippet, usesShared);
+            insertDirectionalLightProcessorArgs(snippet, usesShared);
+            insertFragmentMainArgs(snippet, usesShared);
+            insertPostProcessorArgs(snippet, usesShared);
         }
         fragment() << snippet;
     }
@@ -355,49 +389,61 @@ void QSSGMaterialVertexPipeline::doGenerateWorldNormal(const QSSGShaderDefaultMa
     if (!usesInstancing)
         vertexGenerator.addUniform("qt_normalMatrix", "mat3");
     if (m_hasMorphing)
-        vertexGenerator.append("    qt_vertNormal = qt_getMorphNormal(qt_vertNormal);");
+        vertexGenerator.append("    qt_vertNormal = qt_getTargetNormal(qt_vertNormal);");
     if (m_hasSkinning) {
         vertexGenerator.append("    if (qt_vertWeights != vec4(0.0))");
         vertexGenerator.append("        qt_vertNormal = qt_getSkinNormalMatrix(qt_vertJoints, qt_vertWeights) * qt_vertNormal;");
     }
-    if (!usesInstancing)
-        vertexGenerator.append("    vec3 qt_world_normal = normalize(qt_normalMatrix * qt_vertNormal);");
-    else
+    // If new model->skin is used,
+    // both qt_normalMatrix and qt_modelMatrix are identity.
+    if (!usesInstancing) {
+        if (m_hasSkinning)
+            vertexGenerator.append("    vec3 qt_world_normal = normalize(qt_vertNormal);");
+        else
+            vertexGenerator.append("    vec3 qt_world_normal = normalize(qt_normalMatrix * qt_vertNormal);");
+    } else {
         vertexGenerator.append("    vec3 qt_world_normal = normalize(qt_instancedNormalMatrix * qt_vertNormal);");
+    }
     vertexGenerator.append("    qt_varNormal = qt_world_normal;");
 }
 
 void QSSGMaterialVertexPipeline::doGenerateVarTangent(const QSSGShaderDefaultMaterialKey &inKey)
 {
     if (m_hasMorphing)
-        vertex() << "    qt_vertTangent = qt_getMorphTangent(qt_vertTangent);\n";
+        vertex() << "    qt_vertTangent = qt_getTargetTangent(qt_vertTangent);\n";
     if (m_hasSkinning) {
-        vertex() << "    if (qt_vertWeights != vec4(0.0)) {\n"
-                 << "       qt_vertTangent = (qt_getSkinMatrix(qt_vertJoints, qt_vertWeights) * vec4(qt_vertTangent, 0.0)).xyz;\n"
-                 << "    }\n";
+        vertex() << "    if (qt_vertWeights != vec4(0.0))\n"
+                 << "       qt_vertTangent = (skinMat * vec4(qt_vertTangent, 0.0)).xyz;\n";
 
     }
     const bool usesInstancing = defaultMaterialShaderKeyProperties.m_usesInstancing.getValue(inKey);
-    if (!usesInstancing)
-        vertex() << "    qt_varTangent = (qt_modelMatrix * vec4(qt_vertTangent, 0.0)).xyz;\n";
-    else
+    if (!usesInstancing) {
+        if (!m_hasSkinning)
+            vertex() << "    qt_varTangent = (qt_modelMatrix * vec4(qt_vertTangent, 0.0)).xyz;\n";
+        else
+            vertex() << "    qt_varTangent = qt_vertTangent;\n";
+    } else {
         vertex() << "    qt_varTangent = (qt_instancedModelMatrix * vec4(qt_vertTangent, 0.0)).xyz;\n";
+    }
 }
 
 void QSSGMaterialVertexPipeline::doGenerateVarBinormal(const QSSGShaderDefaultMaterialKey &inKey)
 {
     if (m_hasMorphing)
-        vertex() << "    qt_vertBinormal = qt_getMorphBinormal(qt_vertBinormal);\n";
+        vertex() << "    qt_vertBinormal = qt_getTargetBinormal(qt_vertBinormal);\n";
     if (m_hasSkinning) {
-        vertex() << "    if (qt_vertWeights != vec4(0.0)) {\n"
-                 << "       qt_vertBinormal = (qt_getSkinMatrix(qt_vertJoints, qt_vertWeights) * vec4(qt_vertBinormal, 0.0)).xyz;\n"
-                 << "    }\n";
+        vertex() << "    if (qt_vertWeights != vec4(0.0))\n"
+                 << "       qt_vertBinormal = (skinMat * vec4(qt_vertBinormal, 0.0)).xyz;\n";
     }
     const bool usesInstancing = defaultMaterialShaderKeyProperties.m_usesInstancing.getValue(inKey);
-    if (!usesInstancing)
-        vertex() << "    qt_varBinormal = (qt_modelMatrix * vec4(qt_vertBinormal, 0.0)).xyz;\n";
-    else
+    if (!usesInstancing) {
+        if (!m_hasSkinning)
+            vertex() << "    qt_varBinormal = (qt_modelMatrix * vec4(qt_vertBinormal, 0.0)).xyz;\n";
+        else
+            vertex() << "    qt_varBinormal = qt_vertBinormal;\n";
+    } else {
         vertex() << "    qt_varBinormal = (qt_instancedModelMatrix * vec4(qt_vertBinormal, 0.0)).xyz;\n";
+    }
 }
 
 bool QSSGMaterialVertexPipeline::hasAttributeInKey(QSSGShaderKeyVertexAttribute::VertexAttributeBits inAttr,
